#include "Cpu.h"
#include "display.h"
#include "state.h"

void s_countdown(void);
void s_getid(void);
void s_mainmenu(void);
void writeflash2(unsigned char *addr, unsigned char val);

const char bad9e_str[] = "BAD9E";
const char eor_str[] = "EOR";
const char neg_str[] = "NEG";
const char bit_str[] = "BIT";
const char jsr_str[] = "JSR";
const char sub_str[] = "SUB";
const char add_str[] = "ADD";
const char adc_str[] = "ADC";
const char lda_str[] = "LDA";
const char sta_str[] = "STA";
const char sbc_str[] = "SBC";
const char cpx_str[] = "CPX";
const char cmp_str[] = "CMP";
const char and_str[] = "AND";
const char ldhx_str[] = "LDHX";
const char ora_str[] = "ORA";
const char jmp_str[] = "JMP";
const char simonwinmsg[] = "CHECKBADGECLUESTHEREDON'TCHEATPLAY FAIR ";

const struct simplemenu mainmenu[] = {
    { "POWER", STATE_POWERMENU },
    { "TEST", STATE_TEST },
    { "DEBUG", STATE_DEBUGMENU },
    { "GETID", STATE_GETID_VIEW },
    { "SIMON", STATE_SIMON_START }, 
    { "INFO", STATE_INFO_VIEW },
    { (void *)0, 0 }
};

const struct simplemenu powermenu[] = {
    { "POWERHI", STATE_POWERMENU_HIGH },
    { "POWERLO", STATE_POWERMENU_LOW },
    { (void *)0, 0 }
};

const struct simplemenu debugmenu[] = {
    { "DUMP", STATE_DEBUG_DUMP },
    { "WRITE", STATE_DEBUG_WRITE },
    { "DIS", STATE_DEBUG_DISASSEMBLE },
    { jmp_str, STATE_DEBUG_JUMP },
    { (void *)0, 0 }
};

unsigned short debugptr = 0xfffe;

extern char badgeid[10];
unsigned char curstate = STATE_COUNTDOWN;
extern char curkeys;
typedef union {
    unsigned short sr;
    struct {
        byte sr0 : 8;
        byte sr1 : 8;
    } bytes;
} stateram_t;
#define STATERAM0 stateram.bytes.sr0
#define STATERAM1 stateram.bytes.sr1
stateram_t stateram;
void state(void) {
    switch(curstate) {
        case STATE_COUNTDOWN:
            s_countdown();
            break;
        case STATE_MAINMENU:
            simplemenu(mainmenu, 0);
            break;
        case STATE_GETID_VIEW:
            s_getid();
            break;
        case STATE_INFO_VIEW:
            s_getinfo();
            break;
        case STATE_POWERMENU:
            simplemenu(powermenu, 0);
            break;
        case STATE_POWERMENU_HIGH:
            highpower = 1;
            pop_menu(2);
            break;
        case STATE_POWERMENU_LOW:
            highpower = 0;
            pop_menu(2);
            break;
        case STATE_DEBUGMENU:
            simplemenu(debugmenu, 1);
            break;
        case STATE_DEBUG_DUMP:
            s_debugdump();
            break;
        case STATE_DEBUG_WRITE:
            s_debugwrite();
            break;
        case STATE_DEBUG_DISASSEMBLE:
            s_debug_disassemble();
            break;
        case STATE_SIMON_START:
            s_simon_start();
            break;
        case STATE_SIMON_SHOW:
            s_simon_show();
            break;
        case STATE_SIMON_WIN:
            s_simon_win();
            break;
        case STATE_SIMON_WRONG:
            s_simon_wrong();
            break;
        case STATE_SIMON_FINAL:
            s_simon_final();
            break;
        case STATE_TEST:
            s_test();
            break;
        case STATE_DEBUG_JUMP:
            asm {
                ldhx debugptr
                pshx
                pshh
                rts
            }   
            pop_menu(1);
            break;
    }
}

void stateshift(char newstate) {
    curstate = newstate;
    stateram.sr = 0;
    curkeys = 0;
    state();
}

void s_test(void) {
    unsigned char i;
    if(curkeys & 0x8) {
        pop_menu(1);
        return;
    }
    for(i = 0; i < 10; i++) {
        fbuf[i] = 0xffff;
    }
}
const char info[] = "WWW.XEROBANK.COM/SECRET  ";
void s_getinfo(void) {
    unsigned char i;
    if(!STATERAM0) {
        STATERAM0 = 1;
        STATERAM1 = 0;
    }
    if(curkeys & 0x1) {
        STATERAM1++;
        if(STATERAM1 >= 4) {
            STATERAM1 = 0;
        }
    }
    if(curkeys & 0x8) {
        pop_menu(1);
        return;
    }
    for(i = 0; i < 10; i++) {
        fbuf[i] = LETTER(info[(STATERAM1 * 5) + i]);
    }
}

void s_getid(void) {
    char i;
    if(curkeys & 0x8) {
        pop_menu(1);
        state();
        return;
    }
    for(i = 0; i < 10; i++) {
        fbuf[i] = LETTER(badgeid[i]);
    }
}

char bcount[2];
void s_countdown(void) {
    display_refresh_fbuf();
    if(curkeys & 0x1) {
        stateshift(STATE_MAINMENU);
        return;
    }
    STATERAM0 |= (curkeys & 0x6);
    if(curkeys & 0x2) {
        bcount[0] = ((bcount[0] + 1) & 0xf);
    }
    if(curkeys & 0x4) {
        bcount[1] = ((bcount[1] + 1) & 0xf);
    }
//    if(!STATERAM0) {
//        fbuf[8] = LETTER('T');
//        fbuf[9] = LETTER('Y');
//    } else {
//        fbuf[8] = LETTER(VAL2CHAR(bcount[0]));
//        fbuf[9] = LETTER(VAL2CHAR(bcount[1]));
//    }
}

/* 
 * simple menu.  
 * sr1 is used to hold the current index into the menu.
 * sr0 is used as a flag to reset sr1
 */
void simplemenu(const struct simplemenu * const menu, char showloc) {
    char i, bufoff;
    if(STATERAM0 == 0) {
        STATERAM0 = 1;
        STATERAM1 = 0;
    }
    if(curkeys & 1) {
        push_menu();
        stateshift((char)(menu[STATERAM1].state));
        return;
    }
    if(curkeys & 2) {
        if(STATERAM1 == 0) {
            char x;
            for(x = 0; menu[x+1].name != 0; x++) {
            }
            STATERAM1 = x;
        } else {
            STATERAM1--;
        }
    }
    if(curkeys & 4) {
        STATERAM1++;
        if(menu[STATERAM1].name == 0) {
            STATERAM1 = 0;
        }
    }
    clearfbuf();
    if(showloc) {
        debug_write_address();
        bufoff = 5;
    } else {
        bufoff = 0;
    }
    for(i = 0; i < 10; i++) {
        char c = menu[STATERAM1].name[i];
        if(c == 0) {
            break;
        }
        fbuf[bufoff+i] = LETTER(c);
    }
    if(curkeys & 0x8) {
        pop_menu(1);
        return;
    }
}
struct menustack_entry {
    char state;
    stateram_t stateram;
} menustack[5];
unsigned char menusp = 4;
void push_menu(void) {
    menustack[menusp].state = curstate;
    menustack[menusp].stateram = stateram;
    menusp--;
    menusp &= 0x7;
}
void pop_menu(char num) {
    menusp += num;
    if(menusp & 0xf8 || menustack[menusp].state == curstate) {
        menusp = 7;
        stateshift(STATE_COUNTDOWN);
        return;
    }
    curstate = menustack[menusp].state;
    stateram = menustack[menusp].stateram;
    curkeys = 0;
    state();
}

void debug_write_address(void) {
    fbuf[0] = LETTER('-');
    fbuf[1] = LETTER(VAL2CHAR((unsigned char)(debugptr >> 12)));
    fbuf[2] = LETTER(VAL2CHAR((unsigned char)((debugptr >> 8) & 0xf)));
    fbuf[3] = LETTER(VAL2CHAR((unsigned char)((debugptr >> 4) & 0xf)));
    fbuf[4] = LETTER(VAL2CHAR((unsigned char)(debugptr & 0xf)));
}

void check_debugptr(void) {
    if(debugptr < 0x1800 && debugptr > 0x25e) {
        debugptr = 0x25e;
    }
    if(debugptr > 0x184e && debugptr < 0xe000) {
        debugptr = 0x184e;
    }
}
void debugptr_major_inc(void) {
    unsigned char c = (debugptr >> 8) & 0xff;
    if(c < 0x2) {
        c++;
    } else if(c < 0x18) {
        c = 0x18;
    } else if(c < 0xe0) {
        c = 0xe0;
    } else {
        c++;
    }
    debugptr = (((unsigned short)c) << 8) | (debugptr & 0xff);
    check_debugptr();
}
void s_debugwrite(void) {
    char *ptr = (char *)debugptr;
    if(STATERAM0 == 0) {
        STATERAM0 = 1;
        STATERAM1 = ptr[0];
    }
    if(curkeys & 1) {
        if(debugptr >= 0xe000) {
            writeflash2((unsigned char *)debugptr, STATERAM1);
        } else {
            *(unsigned char *)(debugptr) = STATERAM1;
        }
        debugptr++;
        if(debugptr == 0x260) {
            debugptr = 0x1800;
        } else if(debugptr == 0x1850) {
            debugptr = 0xe000;
        }
        ptr = (char *)debugptr;
        STATERAM1 = ptr[0];
    }
    if(curkeys & 2) {
        STATERAM1 += 0x10;
    }
    if(curkeys & 4) {
        STATERAM1 = (STATERAM1 & 0xf0) | (((STATERAM1 & 0xf) + 1) & 0xf);
    }
    if(curkeys & 8) {
        pop_menu(1);
        return;
    }
    debug_write_address();
    fbuf[5] = LETTER(VAL2CHAR(ptr[0] >> 4));
    fbuf[6] = LETTER(VAL2CHAR(ptr[0] & 0xf));
    fbuf[7] = 0;
    fbuf[8] = LETTER(VAL2CHAR(STATERAM1 >> 4));
    fbuf[9] = LETTER(VAL2CHAR(STATERAM1 & 0xf));
}

void s_debugdump(void) {
    char *ptr;
    if(curkeys & 2) {
        debugptr -= 2;
        check_debugptr();
    }
    if(curkeys & 4) {
        debugptr += 2;
        if(debugptr == 0x25f || debugptr == 0x260) {
            debugptr = 0x1800;
        } else if(debugptr == 0x184f || debugptr == 0x1850) {
            debugptr = 0xe000;
        }
        check_debugptr();
    }
    if(curkeys & 1) {
        debugptr_major_inc();
    }
    if(curkeys & 8) {
        pop_menu(1);
        return;
    }
    ptr = (char *)debugptr;
    debug_write_address();
    fbuf[5] = LETTER(VAL2CHAR(ptr[0] >> 4));
    fbuf[6] = LETTER(VAL2CHAR(ptr[0] & 0xf));
    fbuf[7] = 0;
    fbuf[8] = LETTER(VAL2CHAR(ptr[1] >> 4));
    fbuf[9] = LETTER(VAL2CHAR(ptr[1] & 0xf));
}


struct insn {
    const char * const name;
    const unsigned char nargs;
    const unsigned char args;
};
const struct insn insns_9e60[] = {
    /* 9E60 */ { neg_str, 2, FIRST_FF | SECOND_SP },
    /* 9E61 */ { "CBEQ", 3, FFRR_COMBINED | SECOND_SP },
    /* 9E62 */ { bad9e_str, 0 },
    /* 9E63 */ { "COM", 3, FIRST_FF | SECOND_SP },
    /* 9E64 */ { "LSR", 3, FIRST_FF | SECOND_SP },
    /* 9E65 */ { bad9e_str, 0 },
    /* 9E66 */ { "ROR", 2, FIRST_FF | SECOND_SP },
    /* 9E67 */ { "ASR", 2, FIRST_FF | SECOND_SP },
    /* 9E68 */ { "ASL", 2, FIRST_FF | SECOND_SP },
    /* 9E69 */ { "ROL", 2, FIRST_FF | SECOND_SP },
    /* 9E6A */ { "DEC", 2, FIRST_FF | SECOND_SP },
    /* 9E6B */ { "DBNZ", 3, FFRR_COMBINED | SECOND_SP },
    /* 9E6C */ { "INC", 2, FIRST_FF | SECOND_SP },
    /* 9E6D */ { "TST", 2, FIRST_FF | SECOND_SP },
    /* 9E6E */ { bad9e_str, 0 },
    /* 9E6F */ { "CLR", 2, FIRST_FF | SECOND_SP },
};

const struct insn insns_9ead[] = {
    /* 9EAD */ { bad9e_str, 0 },
    /* 9EAE */ { ldhx_str, 1, FIRST_X },
    /* 9EAF */ { bad9e_str, 0 },

    /* 9EB0 */ { bad9e_str, 0 },
    /* 9EB1 */ { bad9e_str, 0 },
    /* 9EB2 */ { bad9e_str, 0 },
    /* 9EB3 */ { bad9e_str, 0 },
    /* 9EB4 */ { bad9e_str, 0 },
    /* 9EB5 */ { bad9e_str, 0 },
    /* 9EB6 */ { bad9e_str, 0 },
    /* 9EB7 */ { bad9e_str, 0 },
    /* 9EB8 */ { bad9e_str, 0 },
    /* 9EB9 */ { bad9e_str, 0 },
    /* 9EBA */ { bad9e_str, 0 },
    /* 9EBB */ { bad9e_str, 0 },
    /* 9EBC */ { bad9e_str, 0 },
    /* 9EBD */ { bad9e_str, 0 },
    /* 9EBE */ { ldhx_str, 3, EEFF_COMBINED | SECOND_X },
    /* 9EBF */ { bad9e_str, 0 },

    /* 9EC0 */ { bad9e_str, 0 },
    /* 9EC1 */ { bad9e_str, 0 },
    /* 9EC2 */ { bad9e_str, 0 },
    /* 9EC3 */ { bad9e_str, 0 },
    /* 9EC4 */ { bad9e_str, 0 },
    /* 9EC5 */ { bad9e_str, 0 },
    /* 9EC6 */ { bad9e_str, 0 },
    /* 9EC7 */ { bad9e_str, 0 },
    /* 9EC8 */ { bad9e_str, 0 },
    /* 9EC9 */ { bad9e_str, 0 },
    /* 9ECA */ { bad9e_str, 0 },
    /* 9ECB */ { bad9e_str, 0 },
    /* 9ECC */ { bad9e_str, 0 },
    /* 9ECD */ { bad9e_str, 0 },
    /* 9ECE */ { ldhx_str, 2, FIRST_FF | SECOND_X },
    /* 9ECF */ { bad9e_str, 0 },

    /* 9ED0 */ { sub_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9ED1 */ { cmp_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9ED2 */ { sbc_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9ED3 */ { cpx_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9ED4 */ { and_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9ED5 */ { bit_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9ED6 */ { lda_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9ED7 */ { sta_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9ED8 */ { eor_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9ED9 */ { adc_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9EDA */ { ora_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9EDB */ { add_str, 3, EEFF_COMBINED | SECOND_SP },
    /* 9EDC */ { bad9e_str, 0 },
    /* 9EDD */ { bad9e_str, 0 },
    /* 9EDE */ { "LDX", 3, EEFF_COMBINED | SECOND_SP },
    /* 9EDF */ { "STX", 3, EEFF_COMBINED | SECOND_SP },

    /* 9EE0 */ { sub_str, 2, FIRST_FF | SECOND_SP },
    /* 9EE1 */ { cmp_str, 2, FIRST_FF | SECOND_SP },
    /* 9EE2 */ { sbc_str, 2, FIRST_FF | SECOND_SP },
    /* 9EE3 */ { cpx_str, 2, FIRST_FF | SECOND_SP },
    /* 9EE4 */ { and_str, 2, FIRST_FF | SECOND_SP },
    /* 9EE5 */ { bit_str, 2, FIRST_FF | SECOND_SP },
    /* 9EE6 */ { lda_str, 2, FIRST_FF | SECOND_SP },
    /* 9EE7 */ { sta_str, 2, FIRST_FF | SECOND_SP },
    /* 9EE8 */ { eor_str, 2, FIRST_FF | SECOND_SP },
    /* 9EE9 */ { adc_str, 2, FIRST_FF | SECOND_SP },
    /* 9EEA */ { ora_str, 2, FIRST_FF | SECOND_SP },
    /* 9EEB */ { add_str, 2, FIRST_FF | SECOND_SP },
    /* 9EEC */ { bad9e_str, 0 },
    /* 9EED */ { bad9e_str, 0 },
    /* 9EEE */ { "LDX", 2, FIRST_FF | SECOND_SP },
    /* 9EEF */ { "STX", 2, FIRST_FF | SECOND_SP },

    /* 9EF0 */ { bad9e_str, 0 },
    /* 9EF1 */ { bad9e_str, 0 },
    /* 9EF2 */ { bad9e_str, 0 },
    /* 9EF3 */ { "CPHX", 2, FIRST_FF | SECOND_SP },
    /* 9EF4 */ { bad9e_str, 0 },
    /* 9EF5 */ { bad9e_str, 0 },
    /* 9EF6 */ { bad9e_str, 0 },
    /* 9EF7 */ { bad9e_str, 0 },
    /* 9EF8 */ { bad9e_str, 0 },
    /* 9EF9 */ { bad9e_str, 0 },
    /* 9EFA */ { bad9e_str, 0 },
    /* 9EFB */ { bad9e_str, 0 },
    /* 9EFC */ { bad9e_str, 0 },
    /* 9EFD */ { bad9e_str, 0 },
    /* 9EFE */ { ldhx_str, 2, FIRST_FF | SECOND_SP },
    /* 9EFF */ { "STHX", 2, FIRST_FF | SECOND_SP },
};
const struct insn insns[] = {
    /* 00 */ { "BRSET0", 2, FIRST_DD | SECOND_RR }, 
    /* 01 */ { "BRCLR0", 2, FIRST_DD | SECOND_RR }, 
    /* 02 */ { "BRSET1", 2, FIRST_DD | SECOND_RR }, 
    /* 03 */ { "BRCLR1", 2, FIRST_DD | SECOND_RR }, 
    /* 04 */ { "BRSET2", 2, FIRST_DD | SECOND_RR }, 
    /* 05 */ { "BRCLR2", 2, FIRST_DD | SECOND_RR }, 
    /* 06 */ { "BRSET3", 2, FIRST_DD | SECOND_RR }, 
    /* 07 */ { "BRCLR3", 2, FIRST_DD | SECOND_RR }, 
    /* 08 */ { "BRSET4", 2, FIRST_DD | SECOND_RR }, 
    /* 09 */ { "BRCLR4", 2, FIRST_DD | SECOND_RR }, 
    /* 0A */ { "BRSET5", 2, FIRST_DD | SECOND_RR }, 
    /* 0B */ { "BRCLR5", 2, FIRST_DD | SECOND_RR }, 
    /* 0C */ { "BRSET6", 2, FIRST_DD | SECOND_RR }, 
    /* 0D */ { "BRCLR6", 2, FIRST_DD | SECOND_RR }, 
    /* 0E */ { "BRSET7", 2, FIRST_DD | SECOND_RR }, 
    /* 0F */ { "BRCLR7", 2, FIRST_DD | SECOND_RR }, 

    /* 10 */ { "BSET 0", 1, FIRST_DD }, 
    /* 11 */ { "BCLR 0", 1, FIRST_DD }, 
    /* 12 */ { "BSET 1", 1, FIRST_DD }, 
    /* 13 */ { "BCLR 1", 1, FIRST_DD }, 
    /* 14 */ { "BSET 2", 1, FIRST_DD }, 
    /* 15 */ { "BCLR 2", 1, FIRST_DD }, 
    /* 16 */ { "BSET 3", 1, FIRST_DD }, 
    /* 17 */ { "BCLR 3", 1, FIRST_DD }, 
    /* 18 */ { "BSET 4", 1, FIRST_DD }, 
    /* 19 */ { "BCLR 4", 1, FIRST_DD }, 
    /* 1A */ { "BSET 5", 1, FIRST_DD }, 
    /* 1B */ { "BCLR 5", 1, FIRST_DD }, 
    /* 1C */ { "BSET 6", 1, FIRST_DD }, 
    /* 1D */ { "BCLR 6", 1, FIRST_DD }, 
    /* 1E */ { "BSET 7", 1, FIRST_DD }, 
    /* 1F */ { "BCLR 7", 1, FIRST_DD }, 

    /* 20 */ { "BRA", 1, FIRST_RR },
    /* 21 */ { "BRN", 1, FIRST_RR },
    /* 22 */ { "BHI", 1, FIRST_RR },
    /* 23 */ { "BLS", 1, FIRST_RR },
    /* 24 */ { "BHS", 1, FIRST_RR },
    /* 25 */ { "BLO", 1, FIRST_RR },
    /* 26 */ { "BNE", 1, FIRST_RR },
    /* 27 */ { "BEQ", 1, FIRST_RR },
    /* 28 */ { "BHCC", 1, FIRST_RR },
    /* 29 */ { "BHCS", 1, FIRST_RR },
    /* 2A */ { "BPL", 1, FIRST_RR },
    /* 2B */ { "BMI", 1, FIRST_RR },
    /* 2C */ { "BMC", 1, FIRST_RR },
    /* 2D */ { "BMS", 1, FIRST_RR },
    /* 2E */ { "BIL", 1, FIRST_RR },
    /* 2F */ { "BIH", 1, FIRST_RR },

    /* 30 */ { neg_str, 1, FIRST_DD },
    /* 31 */ { "CBEQ", 2, FIRST_DD | SECOND_RR },
    /* 32 */ { ldhx_str, 2, HHLL_COMBINED },
    /* 33 */ { "COM", 1, FIRST_DD },
    /* 34 */ { "LSR", 1, FIRST_DD },
    /* 35 */ { "STHX", 1, FIRST_DD },
    /* 36 */ { "ROR", 1, FIRST_DD },
    /* 37 */ { "ASR", 1, FIRST_DD },
    /* 38 */ { "ASL", 1, FIRST_DD },
    /* 39 */ { "ROL", 1, FIRST_DD },
    /* 3A */ { "DEC", 1, FIRST_DD },
    /* 3B */ { "DBNZ", 2, FIRST_DD | SECOND_RR },
    /* 3C */ { "INC", 1, FIRST_DD },
    /* 3D */ { "TST", 1, FIRST_DD },
    /* 3E */ { "STHX", 2, HHLL_COMBINED },
    /* 3F */ { "CLR", 1, FIRST_DD },

    /* 40 */ { "NEGA", 0 },
    /* 41 */ { "CBEQA", 2, FIRST_II | SECOND_RR },
    /* 42 */ { "MUL", 0 },
    /* 43 */ { "COMA", 0 },
    /* 44 */ { "LSRA", 0 },
    /* 45 */ { "LDHX#", 2, JJKK_COMBINED },
    /* 46 */ { "RORA", 0 },
    /* 47 */ { "ASRA", 0 },
    /* 48 */ { "ASLA", 0 },
    /* 49 */ { "ROLA", 0 },
    /* 4A */ { "DECA", 0 },
    /* 4B */ { "DBNZA", 1, FIRST_RR },
    /* 4C */ { "INCA", 0 },
    /* 4D */ { "TSTA", 0 },
    /* 4E */ { "MOV", 2, FIRST_DD | SECOND_DD },
    /* 4F */ { "CLRA", 0 },


    /* 50 */ { "NEGX", 0 },
    /* 51 */ { "CBEQX", 2, FIRST_II | SECOND_RR },
    /* 52 */ { "DIV", 0 },
    /* 53 */ { "COMX", 0 },
    /* 54 */ { "LSRX", 0 },
    /* 55 */ { ldhx_str, 1, FIRST_DD },
    /* 56 */ { "RORX", 0 },
    /* 57 */ { "ASRX", 0 },
    /* 58 */ { "ASLX", 0 },
    /* 59 */ { "ROLX", 0 },
    /* 5A */ { "DECX", 0 },
    /* 5B */ { "DBNZX", 1, FIRST_RR },
    /* 5C */ { "INCX", 0 },
    /* 5D */ { "TSTX", 0 },
    /* 5E */ { "MOV", 2, FIRST_DD | SECOND_XPLUS },
    /* 5F */ { "CLRX", 0 },

    /* 60 */ { neg_str, 2, FIRST_II | SECOND_X },
    /* 61 */ { "CBEQ", 3, FFRR_COMBINED | SECOND_X },
    /* 62 */ { "NSA", 0 },
    /* 63 */ { "COM", 2, FIRST_II | SECOND_X },
    /* 64 */ { "LSR", 2, FIRST_II | SECOND_X },
    /* 65 */ { "CPHX#", 2, JJKK_COMBINED },
    /* 66 */ { "ROR", 2, FIRST_II | SECOND_X },
    /* 67 */ { "ASR", 2, FIRST_II | SECOND_X },
    /* 68 */ { "ASL", 2, FIRST_II | SECOND_X },
    /* 69 */ { "ROL", 2, FIRST_II | SECOND_X },
    /* 6A */ { "DEC", 2, FIRST_II | SECOND_X },
    /* 6B */ { "DBNZ", 3, FFRR_COMBINED | SECOND_X },
    /* 6C */ { "INC", 2, FIRST_II | SECOND_X },
    /* 6D */ { "TST", 2, FIRST_II | SECOND_X },
    /* 6E */ { "MOV#", 2, FIRST_II | SECOND_DD },
    /* 6F */ { "CLR", 2, FIRST_II | SECOND_X },

    /* 70 */ { neg_str, 1, FIRST_X },
    /* 71 */ { "CBEQ ,X+", 1, FIRST_RR },
    /* 72 */ { "DAA", 0 },
    /* 73 */ { "COM", 1, FIRST_X },
    /* 74 */ { "LSR", 1, FIRST_X },
    /* 75 */ { "CPHX", 1, FIRST_DD },
    /* 76 */ { "ROR", 1, FIRST_X },
    /* 77 */ { "ASR", 1, FIRST_X },
    /* 78 */ { "ASL", 1, FIRST_X },
    /* 79 */ { "ROL", 1, FIRST_X },
    /* 7A */ { "DEC", 1, FIRST_X },
    /* 7B */ { "DBNZ", 2, FIRST_X | SECOND_RR },
    /* 7C */ { "INC", 1, FIRST_X },
    /* 7D */ { "TST", 1, FIRST_X },
    /* 7E */ { "MOV", 2, FIRST_XPLUS | SECOND_DD },
    /* 7F */ { "CLR", 1, FIRST_X },

    /* 80 */ { "RTI", 0 },
    /* 81 */ { "RTS", 0 },
    /* 82 */ { "BGND", 0 },
    /* 83 */ { "SWI", 0 },
    /* 84 */ { "TAP", 0 },
    /* 85 */ { "TPA", 0 },
    /* 86 */ { "PULA", 0 },
    /* 87 */ { "PSHA", 0 },
    /* 88 */ { "PULX", 0 },
    /* 89 */ { "PSHX", 0 },
    /* 8A */ { "PULH", 0 },
    /* 8B */ { "PSHH", 0 },
    /* 8C */ { "CLRH", 0 },
    /* 8D */ { "BAD8D", 0 },
    /* 8E */ { "STOP", 0 },
    /* 8F */ { "WAIT", 0 },

    /* 90 */ { "BGE", 1, FIRST_RR },
    /* 91 */ { "BLT", 1, FIRST_RR },
    /* 92 */ { "BGT", 1, FIRST_RR },
    /* 93 */ { "BLE", 1, FIRST_RR },
    /* 94 */ { "TXS", 0 },
    /* 95 */ { "TSX", 0 },
    /* 96 */ { "STHX", 2, HHLL_COMBINED },
    /* 97 */ { "TAX", 0 },
    /* 98 */ { "CLC", 0 },
    /* 99 */ { "SEC", 0 },
    /* 9A */ { "CLI", 0 },
    /* 9B */ { "SEI", 0 },
    /* 9C */ { "RSP", 0 },
    /* 9D */ { "NOP", 0 },
    /* 9E */ { "9E", 0 },
    /* 9F */ { "TXA", 0 },

    /* A0 */ { sub_str, 1, FIRST_II },
    /* A1 */ { cmp_str, 1, FIRST_II },
    /* A2 */ { sbc_str, 1, FIRST_II },
    /* A3 */ { cpx_str, 1, FIRST_II },
    /* A4 */ { and_str, 1, FIRST_II },
    /* A5 */ { bit_str, 1, FIRST_II },
    /* A6 */ { lda_str, 1, FIRST_II },
    /* A7 */ { "AIS", 1, FIRST_II },
    /* A8 */ { eor_str, 1, FIRST_II },
    /* A9 */ { adc_str, 1, FIRST_II },
    /* AA */ { ora_str, 1, FIRST_II },
    /* AB */ { add_str, 1, FIRST_II },
    /* AC */ { "BADAC", 0 },
    /* AD */ { "BSR", 1, FIRST_RR },
    /* AE */ { "LDX", 1, FIRST_II },
    /* AF */ { "AIX", 1, FIRST_II },

    /* B0 */ { sub_str, 1, FIRST_DD },
    /* B1 */ { cmp_str, 1, FIRST_DD },
    /* B2 */ { sbc_str, 1, FIRST_DD },
    /* B3 */ { cpx_str, 1, FIRST_DD },
    /* B4 */ { and_str, 1, FIRST_DD },
    /* B5 */ { bit_str, 1, FIRST_DD },
    /* B6 */ { lda_str, 1, FIRST_DD },
    /* B7 */ { sta_str, 1, FIRST_DD },
    /* B8 */ { eor_str, 1, FIRST_DD },
    /* B9 */ { adc_str, 1, FIRST_DD },
    /* BA */ { ora_str, 1, FIRST_DD },
    /* BB */ { add_str, 1, FIRST_DD },
    /* BC */ { jmp_str, 1, FIRST_DD },
    /* BD */ { jsr_str, 1, FIRST_DD },
    /* BE */ { "LDX", 1, FIRST_DD },
    /* BF */ { "STX", 1, FIRST_DD },

    /* C0 */ { sub_str, 2, HHLL_COMBINED },
    /* C1 */ { cmp_str, 2, HHLL_COMBINED },
    /* C2 */ { sbc_str, 2, HHLL_COMBINED },
    /* C3 */ { cpx_str, 2, HHLL_COMBINED },
    /* C4 */ { and_str, 2, HHLL_COMBINED },
    /* C5 */ { bit_str, 2, HHLL_COMBINED },
    /* C6 */ { lda_str, 2, HHLL_COMBINED },
    /* C7 */ { sta_str, 2, HHLL_COMBINED },
    /* C8 */ { eor_str, 2, HHLL_COMBINED },
    /* C9 */ { adc_str, 2, HHLL_COMBINED },
    /* CA */ { ora_str, 2, HHLL_COMBINED },
    /* CB */ { add_str, 2, HHLL_COMBINED },
    /* CC */ { jmp_str, 2, HHLL_COMBINED },
    /* CD */ { jsr_str, 2, HHLL_COMBINED },
    /* CE */ { "LDX", 2, HHLL_COMBINED },
    /* CF */ { "STX", 2, HHLL_COMBINED },

    /* D0 */ { sub_str, 3, EEFF_COMBINED | SECOND_X },
    /* D1 */ { cmp_str, 3, EEFF_COMBINED | SECOND_X },
    /* D2 */ { sbc_str, 3, EEFF_COMBINED | SECOND_X },
    /* D3 */ { cpx_str, 3, EEFF_COMBINED | SECOND_X },
    /* D4 */ { and_str, 3, EEFF_COMBINED | SECOND_X },
    /* D5 */ { bit_str, 3, EEFF_COMBINED | SECOND_X },
    /* D6 */ { lda_str, 3, EEFF_COMBINED | SECOND_X },
    /* D7 */ { sta_str, 3, EEFF_COMBINED | SECOND_X },
    /* D8 */ { eor_str, 3, EEFF_COMBINED | SECOND_X },
    /* D9 */ { adc_str, 3, EEFF_COMBINED | SECOND_X },
    /* DA */ { ora_str, 3, EEFF_COMBINED | SECOND_X },
    /* DB */ { add_str, 3, EEFF_COMBINED | SECOND_X },
    /* DC */ { jmp_str, 3, EEFF_COMBINED | SECOND_X },
    /* DD */ { jsr_str, 3, EEFF_COMBINED | SECOND_X },
    /* DE */ { "LDX", 3, EEFF_COMBINED | SECOND_X },
    /* DF */ { "STX", 3, EEFF_COMBINED | SECOND_X },

    /* E0 */ { sub_str, 2, FIRST_DD | SECOND_X },
    /* E1 */ { cmp_str, 2, FIRST_DD | SECOND_X },
    /* E2 */ { sbc_str, 2, FIRST_DD | SECOND_X },
    /* E3 */ { cpx_str, 2, FIRST_DD | SECOND_X },
    /* E4 */ { and_str, 2, FIRST_DD | SECOND_X },
    /* E5 */ { bit_str, 2, FIRST_DD | SECOND_X },
    /* E6 */ { lda_str, 2, FIRST_DD | SECOND_X },
    /* E7 */ { sta_str, 2, FIRST_DD | SECOND_X },
    /* E8 */ { eor_str, 2, FIRST_DD | SECOND_X },
    /* E9 */ { adc_str, 2, FIRST_DD | SECOND_X },
    /* EA */ { ora_str, 2, FIRST_DD | SECOND_X },
    /* EB */ { add_str, 2, FIRST_DD | SECOND_X },
    /* EC */ { jmp_str, 2, FIRST_DD | SECOND_X },
    /* ED */ { jsr_str, 2, FIRST_DD | SECOND_X },
    /* EE */ { "LDX", 2, FIRST_DD | SECOND_X },
    /* EF */ { "STX", 2, FIRST_DD | SECOND_X },

    /* F0 */ { sub_str, 1, FIRST_X },
    /* F1 */ { cmp_str, 1, FIRST_X },
    /* F2 */ { sbc_str, 1, FIRST_X },
    /* F3 */ { cpx_str, 1, FIRST_X },
    /* F4 */ { and_str, 1, FIRST_X },
    /* F5 */ { bit_str, 1, FIRST_X },
    /* F6 */ { lda_str, 1, FIRST_X },
    /* F7 */ { sta_str, 1, FIRST_X },
    /* F8 */ { eor_str, 1, FIRST_X },
    /* F9 */ { adc_str, 1, FIRST_X },
    /* FA */ { ora_str, 1, FIRST_X },
    /* FB */ { add_str, 1, FIRST_X },
    /* FC */ { jmp_str, 1, FIRST_X },
    /* FD */ { jsr_str, 1, FIRST_X },
    /* FE */ { "LDX", 1, FIRST_X },
    /* FF */ { "STX", 1, FIRST_X },
};

const char dummyinsn[] = { 0x30, 0x61, 0x9d, 0x9d, 0x9e, 0x6a, 0x50, 0x9e, 0xff, 0xff, 0x30, 0x61, 0xfe, 0xfd, 0xff, 0xe0, 0xff, 0x9d, 0x9d, 0x09, 0x3C, 0xFD, 0x32, 0xFA, 0xD0, 0x10, 0xEA, 0x09, 0x3C, 0xFD };
union {
    unsigned char debugbuf[25];
    struct {
        unsigned char curval;
    } writestate;
    struct {
        unsigned char simonptr;
        unsigned char simonbuf[24];
    } simondata;
} debugdata;
#define SIMONBUF debugdata.simondata.simonbuf
#define SIMONPTR debugdata.simondata.simonptr
#define DEBUGBUF debugdata.debugbuf
#define SIMON_LEVEL simondata.level
#define SIMON_DATA simondata.data
unsigned char *nextinsn;
void s_debug_disassemble(void) {
    unsigned char *ptr;
    unsigned char i, ni;
    struct insn *in;
    //debugptr = (unsigned short)&dummyinsn;  //XXX remove
    if(curkeys & 4 && STATERAM0) {
        debugptr = (unsigned short)nextinsn;
        check_debugptr();
        STATERAM0 = 0;
    }
    ptr = (unsigned char *)debugptr;

    if(ptr[0] == 0x9e) {
        ptr++;
        if(ptr[0] < 0x60) {
            in = (struct insn *)&insns_9ead[0];
        } else if(ptr[0] < 0x70) {
            in = (struct insn *)&insns_9e60[ptr[0]-0x60];
        } else if(ptr[0] < 0xae) {
            in = (struct insn *)&insns_9ead[0];
        } else {
            in = (struct insn *)&insns_9ead[ptr[0]-0xad];
        }
    } else {
        in = (struct insn *)&insns[ptr[0]];
    }
    if(STATERAM0 == 0) {
        unsigned char nextptr, args, ptroffset = 0;
        unsigned char *insnargs = (unsigned char *)&(in->args);
        STATERAM0 = 1;
        STATERAM1 = 0;
        DEBUGBUF[0] = '-';
        DEBUGBUF[1] = VAL2CHAR((unsigned char)(debugptr >> 12));
        DEBUGBUF[2] = VAL2CHAR((unsigned char)((debugptr >> 8) & 0xf));
        DEBUGBUF[3] = VAL2CHAR((unsigned char)((debugptr >> 4) & 0xf));
        DEBUGBUF[4] = VAL2CHAR((unsigned char)(debugptr & 0xf));
        for(i = 5; i < sizeof(DEBUGBUF); i++) {
            DEBUGBUF[i] = ' ';
        }
        for(i = 0; in->name[i] != 0; i++) {
            DEBUGBUF[5+i] = in->name[i];
        }
        if(i >= 5) {
            nextptr = 15;
        } else {
            nextptr = 10;
        }
        nextinsn = ptr + 1 + (in->nargs);
        for(args = 0; args < in->nargs && args < 2; args++) {
            unsigned char curarg;
            if(args == 0) {
                curarg = ((in->args) >> 4) & 0xf;
            } else {
                curarg = in->args & 0xf;
            }
            if(curarg == ARG_XPLUS || curarg == ARG_X || curarg == ARG_SP) {
                nextinsn--;
                ptroffset++;
            }
            decode_arg(nextptr, args, curarg, (unsigned char)(ptr[args+1-ptroffset]));
            nextptr += 5;
        }
    }
    if(curkeys & 1) {
        STATERAM1++;
        if(STATERAM1 == 4 || DEBUGBUF[5*(STATERAM1+1)] == ' ') {
            STATERAM1 = 0;
        }
    }
    if(curkeys & 8) {
        pop_menu(1);
        return;
    }

    ni = 5 * STATERAM1; 
    for(i = 0; i < 10; i++) {
        fbuf[i] = LETTER(DEBUGBUF[ni+i]);
    }
}
void s_simon_wrong(void) {
    clearfbuf();
    fbuf[0] = LETTER('W');
    fbuf[1] = LETTER('R');
    fbuf[2] = LETTER('O');
    fbuf[3] = LETTER('N');
    fbuf[4] = LETTER('G');
    if(curkeys) {
        pop_menu(1);
        return;
    }
}

void s_simon_win(void) {
    clearfbuf();
    fbuf[1] = LETTER('R');
    fbuf[2] = LETTER('A');
    fbuf[3] = LETTER('D');
    if(curkeys) {
        pop_menu(1);
        return;
    }
}

#define MAXSIMONLEVEL 10
unsigned char simonlevel = 1;
unsigned char simontimer = 0;
void s_simon_start(void) {
    unsigned short nr;
    if(curkeys & 1) {
        push_menu();
        stateshift(STATE_SIMON_SHOW);
        return;
    }
    if(curkeys & 8) {
        pop_menu(1);
        return;
    }
    for(SIMONPTR = 0; SIMONPTR < simonlevel; SIMONPTR++) {
        nr = nextrand() % 3;
        SIMONBUF[SIMONPTR] = (unsigned char)(nr & 3);
    }
    clearfbuf();
    fbuf[0] = LETTER('L');
    fbuf[1] = LETTER('E');
    fbuf[2] = LETTER('V');
    fbuf[3] = LETTER('E');
    fbuf[4] = LETTER('L');
    fbuf[5] = LETTER(VAL2CHAR(simonlevel >> 4));
    fbuf[6] = LETTER(VAL2CHAR(simonlevel & 0xf));
}
void simon_timer(void) {
    if(curstate == STATE_SIMON_SHOW && STATERAM1 == 0) {
        if(simontimer == 0) {
            fbuf[0] = LETTER('0');
            fbuf[1] = LETTER(VAL2CHAR((SIMONBUF[SIMONPTR]) & 0x3));
            SIMONPTR = SIMONPTR + 1;
            if(SIMONPTR > simonlevel) {
                clearfbuf();
                STATERAM1 = 1;
                SIMONPTR = 0;
                simon_show_remaining();
            }
        } else if(simontimer == 10) {
            fbuf[0] = 0;
            fbuf[1] = 0;
        }
        simontimer++;
        if(simontimer >= 15) {
            simontimer = 0;
        }
    }
}

void simon_show_remaining(void) {
    unsigned char c = simonlevel - SIMONPTR;
    fbuf[0] = LETTER('G');
    fbuf[1] = LETTER('O');
    fbuf[5] = LETTER(VAL2CHAR((c) >> 4));
    fbuf[6] = LETTER(VAL2CHAR((c) & 0xf));
}

void s_simon_show(void) {
    char c; 
    if(STATERAM0 == 0) {
        clearfbuf();
        STATERAM0 = 1;
        STATERAM1 = 0;
        SIMONPTR = 0;
        simontimer = 0;
    }
    if(curkeys & 8) {
        pop_menu(2);
        return;
    }
    if(STATERAM1) {
        c = SIMONBUF[SIMONPTR];
        if(curkeys && (1 << c) != curkeys) {
            STATERAM1 = 0;
            SIMONPTR = 0;
            stateshift(STATE_SIMON_WRONG);
            return;
        }
        SIMONPTR++;
        if(SIMONPTR >= simonlevel) {
            simonlevel++;
            if(simonlevel > MAXSIMONLEVEL) {
                simonlevel = MAXSIMONLEVEL;
                stateshift(STATE_SIMON_FINAL);
            } else {
                stateshift(STATE_SIMON_WIN);
            }
            return;
        } 
        simon_show_remaining();
    }
}

void s_simon_final(void) {
    unsigned char i;
    if(!STATERAM0) {
        STATERAM0 = 1;
        STATERAM1 = 0;
    }
    if(curkeys & 0x1) {
        STATERAM1++;
        if(STATERAM1 >= 7) {
            STATERAM1 = 0;
        }
    }
    if(curkeys & 0x8) {
        stateshift(STATE_COUNTDOWN);
        return;
    }
    for(i = 0; i < 10; i++) {
        fbuf[i] = LETTER(simonwinmsg[(STATERAM1 * 5) + i]);
    }
}



void decode_arg(unsigned char bufptr, unsigned char argpos, unsigned char arg, unsigned char argval) {
    unsigned short addr;
    short sx;
    switch(arg) {
        case ARG_XPLUS:
            DEBUGBUF[bufptr+2] = '+';
            // FALLTHROUGH
        case ARG_X:
            DEBUGBUF[bufptr] = ',';
            DEBUGBUF[bufptr+1] = 'X';
            return;
            break;
        case ARG_SP:
            DEBUGBUF[bufptr] = ',';
            DEBUGBUF[bufptr+1] = 'S';
            DEBUGBUF[bufptr+2] = 'P';
            return;
            break;
        case ARG_DD:
            addr = (*(unsigned char *)(&argval));
            break;
        case ARG_II:
            addr = argval;
            break;
        case ARG_RR: 
            sx = (argval & 0x7f);
            if(argval & 0x80) {
                sx = -(0x80-sx);
            }
            addr = ((unsigned short)nextinsn) + sx;
            break; 
        case ARG_EEFF:
            addr = *(unsigned short*)(((unsigned short)nextinsn) - 2);
            if(addr & 0x8000) {
                addr = 0xffff - ~(addr) + 1;
                DEBUGBUF[bufptr] = '-';
            } else {
                DEBUGBUF[bufptr] = '+';
            }
            DEBUGBUF[bufptr+1] = VAL2CHAR((addr >> 12) & 0xf);
            DEBUGBUF[bufptr+2] = VAL2CHAR((addr >> 8) & 0xf);
            DEBUGBUF[bufptr+3] = VAL2CHAR((addr >> 4) & 0xf);
            DEBUGBUF[bufptr+4] = VAL2CHAR(addr & 0xf);
            return;
        case ARG_FFRR:
            if (argpos == 0) {
                DEBUGBUF[bufptr] = VAL2CHAR(argval >> 4);
                DEBUGBUF[bufptr+1] = VAL2CHAR(argval & 0xf);
                argval = *(unsigned char *)(((unsigned short)nextinsn)-2);
                sx = (argval & 0x7f);
                if(argval & 0x80) {
                    sx = -(0x80-sx);
                }
                addr = ((unsigned short)nextinsn-1) + sx;
                bufptr += 10;
            }
            break;
        case ARG_JJKK:
            // FALLTHROUGH
        case ARG_HHLL: 
            if(argpos > 0) {
                return;
            }
            addr = *(unsigned short*)(((unsigned short)nextinsn) - 2);
            break;
    }
    DEBUGBUF[bufptr] = VAL2CHAR((addr >> 12) & 0xf);
    DEBUGBUF[bufptr+1] = VAL2CHAR((addr >> 8) & 0xf);
    DEBUGBUF[bufptr+2] = VAL2CHAR((addr >> 4) & 0xf);
    DEBUGBUF[bufptr+3] = VAL2CHAR(addr & 0xf);
}


