#include "time.h"
#include "Events.h"
#include "PE_Error.h"
#include "LEDTimer.h"
#pragma MESSAGE DISABLE C2705          /* WARNING C2705: Possible loss of data */
#pragma MESSAGE DISABLE C5919          /* WARNING C5919: Conversion of floating to unsigned integral */
#pragma MESSAGE DISABLE C5703          /* WARNING C5703: Parameter X declared in function F but not referenced */
#pragma MESSAGE DISABLE C4002          /* Disable warning C4002 "Result not used" */

/* Internal method prototypes */


/*
** ===================================================================
**     Method      :  LEDTimer_Init (bean TimerInt)
**
**     Description :
**         Initializes the associated peripheral(s) and the beans 
**         internal variables. The method is called automatically as a 
**         part of the application initialization code.
**         This method is internal. It is used by Processor Expert only.
** ===================================================================
*/
void LEDTimer_Init(void)
{
  /* TPM1SC: TOF=0,TOIE=0,CPWMS=0,CLKSB=0,CLKSA=0,PS2=0,PS1=0,PS0=0 */
  setReg8(TPM1SC, 0x00);               /* Stop HW; disable overflow interrupt and set prescaler to 0 */ 
  /* TPM1C0SC: CH0F=0,CH0IE=1,MS0B=0,MS0A=1,ELS0B=0,ELS0A=0,??=0,??=0 */
  setReg8(TPM1C0SC, 0x50);             /* Set output compare mode and enable compare interrupt */ 
  LEDTimer_SetCV(0x1000U);             /* Inicialize appropriate value to the compare/modulo/reload register */
  /* TPM1CNTH: BIT15=0,BIT14=0,BIT13=0,BIT12=0,BIT11=0,BIT10=0,BIT9=0,BIT8=0 */
  setReg8(TPM1CNTH, 0x00);             /* Reset HW Counter */ 
//  /* TPM1SC: TOF=0,TOIE=0,CPWMS=0,CLKSB=0,CLKSA=1,PS2=1,PS1=0,PS0=1 */
//  setReg8(TPM1SC, 0x0D);               /* Set prescaler and run counter */ 
  /* TPM1SC: TOF=0,TOIE=0,CPWMS=0,CLKSB=0,CLKSA=1,PS2=0,PS1=0,PS0=0 */
  setReg8(TPM1SC, 0x08);               /* Set prescaler and run counter */ 

//  /* TPM2SC: TOF=0, TOIE=0, CPWMS=0, CLKSB=0, CLKSA=0, PS2=0, PS1=0, PS0=0 */
//  setReg8(TPM2SC, 0x0);
//  /* 10 = falling edge */
//  /* TPM2C0SC: CH0F=0, CH0IE=1, MS0B=0, MS0A=0, ELS0B=1, ELS0A=0, pin1=0, pin0=0 */
//  setReg8(TPM2C0SC, 0x48);
//  /* TPM2C0SC: CH0F=0, CH0IE=1, MS0B=0, MS0A=1, ELS0B=0, ELS0A=0, pin1=0, pin0=0 */
//  setReg8(TPM2C0SC, 0x50);
//  /* TPM2C0SC: CH0F=0, CH0IE=0, MS0B=0, MS0A=1, ELS0B=0, ELS0A=0, pin1=0, pin0=0 */
//  setReg8(TPM2C0SC, 0x10);
//  TPM2C0V = 0;
//  TPM2MOD = 0xffff;
//  TPM2CNT = 0;
//  /* TPM2SC: TOF=0, TOIE=0, CPWMS=0, CLKSB=0, CLKSA=1, PS2=1, PS1=1, PS0=1 */   /* so 0xffff ~= 2s */
//  setReg8(TPM2SC, 0xf);
//  /* TPM2SC: TOF=0, TOIE=0, CPWMS=0, CLKSB=0, CLKSA=1, PS2=0, PS1=0, PS0=0 */   /* so 0xffff ~= 2s */
//  setReg8(TPM2SC, 0x8);

}


ISR(LEDTimer_Interrupt)
{
    if(TPM1C0SC & 0x80) {
        /* TPM1C0SC: CH0F=0 */
        clrReg8Bits(TPM1C0SC, 0x80);         /* Reset compare interrupt request flag */ 
    }
    LEDTimer_OnInterrupt();              /* Invoke user event */
}

